import json
from typing import Union, Dict, Any

from .structs import (
    PublicKeyCredentialCreationOptions,
    PublicKeyCredentialRequestOptions,
)
from .bytes_to_base64url import bytes_to_base64url


def options_to_json(
    options: Union[
        PublicKeyCredentialCreationOptions,
        PublicKeyCredentialRequestOptions,
    ]
) -> str:
    """
    Prepare options for transmission to the front end as JSON
    """
    if isinstance(options, PublicKeyCredentialCreationOptions):
        _rp = {"name": options.rp.name}
        if options.rp.id:
            _rp["id"] = options.rp.id

        _user: Dict[str, Any] = {
            "id": bytes_to_base64url(options.user.id),
            "name": options.user.name,
            "displayName": options.user.display_name,
        }

        reg_to_return: Dict[str, Any] = {
            "rp": _rp,
            "user": _user,
            "challenge": bytes_to_base64url(options.challenge),
            "pubKeyCredParams": [
                {"type": param.type, "alg": param.alg} for param in options.pub_key_cred_params
            ],
        }

        # Begin handling optional values

        if options.timeout is not None:
            reg_to_return["timeout"] = options.timeout

        if options.exclude_credentials is not None:
            _excluded = options.exclude_credentials
            json_excluded = []

            for cred in _excluded:
                json_excluded_cred: Dict[str, Any] = {
                    "id": bytes_to_base64url(cred.id),
                    "type": cred.type.value,
                }

                if cred.transports:
                    json_excluded_cred["transports"] = [
                        transport.value for transport in cred.transports
                    ]

                json_excluded.append(json_excluded_cred)

            reg_to_return["excludeCredentials"] = json_excluded

        if options.authenticator_selection is not None:
            _selection = options.authenticator_selection
            json_selection: Dict[str, Any] = {}

            if _selection.authenticator_attachment is not None:
                json_selection["authenticatorAttachment"] = (
                    _selection.authenticator_attachment.value
                )

            if _selection.resident_key is not None:
                json_selection["residentKey"] = _selection.resident_key.value

            if _selection.require_resident_key is not None:
                json_selection["requireResidentKey"] = _selection.require_resident_key

            if _selection.user_verification is not None:
                json_selection["userVerification"] = _selection.user_verification.value

            reg_to_return["authenticatorSelection"] = json_selection

        if options.attestation is not None:
            reg_to_return["attestation"] = options.attestation.value

        if options.hints is not None:
            reg_to_return["hints"] = [hint.value for hint in options.hints]

        return json.dumps(reg_to_return)

    if isinstance(options, PublicKeyCredentialRequestOptions):
        auth_to_return: Dict[str, Any] = {"challenge": bytes_to_base64url(options.challenge)}

        if options.timeout is not None:
            auth_to_return["timeout"] = options.timeout

        if options.rp_id is not None:
            auth_to_return["rpId"] = options.rp_id

        if options.allow_credentials is not None:
            _allowed = options.allow_credentials
            json_allowed = []

            for cred in _allowed:
                json_allowed_cred: Dict[str, Any] = {
                    "id": bytes_to_base64url(cred.id),
                    "type": cred.type.value,
                }

                if cred.transports:
                    json_allowed_cred["transports"] = [
                        transport.value for transport in cred.transports
                    ]

                json_allowed.append(json_allowed_cred)

            auth_to_return["allowCredentials"] = json_allowed

        if options.user_verification:
            auth_to_return["userVerification"] = options.user_verification.value

        return json.dumps(auth_to_return)

    raise TypeError(
        "Options was not instance of PublicKeyCredentialCreationOptions or PublicKeyCredentialRequestOptions"
    )



from cryptography.x509 import load_pem_x509_certificate
from OpenSSL.crypto import X509


def pem_cert_bytes_to_open_ssl_x509(cert: bytes) -> X509:
    """Convert PEM-formatted certificate bytes into an X509 instance usable for cert
    chain validation
    """
    cert_crypto = load_pem_x509_certificate(cert)
    cert_openssl = X509().from_cryptography(cert_crypto)
    return cert_openssl



from base64 import urlsafe_b64encode


def bytes_to_base64url(val: bytes) -> str:
    """
    Base64URL-encode the provided bytes
    """
    return urlsafe_b64encode(val).decode("utf-8").rstrip("=")



from unittest import TestCase

from webauthn.helpers.tpm import map_tpm_manufacturer_id


class TestWebAuthnGenerateUserHandle(TestCase):
    def test_handles_recognized_id(self) -> None:
        info = map_tpm_manufacturer_id("id:4353434F")

        self.assertEqual(info.name, "Cisco")
        self.assertEqual(info.id, "CSCO")

    def test_raises_on_unrecognized_id(self) -> None:
        with self.assertRaises(KeyError):
            map_tpm_manufacturer_id("id:FFFFFFFF")



from typing import Any

import cbor2

from .exceptions import InvalidCBORData


def encode_cbor(val: Any) -> bytes:
    """
    Attempt to encode data into CBOR.

    Raises:
        `helpers.exceptions.InvalidCBORData` if data cannot be decoded
    """
    try:
        to_return = cbor2.dumps(val)
    except Exception as exc:
        raise InvalidCBORData("Data could not be encoded to CBOR") from exc

    return to_return



import json
from unittest import TestCase
from unittest.mock import MagicMock, patch

from webauthn.helpers import (
    base64url_to_bytes,
    bytes_to_base64url,
    encode_cbor,
    parse_registration_credential_json,
    parse_cbor,
    parse_attestation_object,
)
from webauthn.helpers.exceptions import InvalidRegistrationResponse, InvalidCBORData
from webauthn.helpers.known_root_certs import globalsign_r2
from webauthn.helpers.structs import (
    AttestationFormat,
    PublicKeyCredentialType,
)
from webauthn import verify_registration_response


class TestVerifyRegistrationResponse(TestCase):
    def test_verifies_none_attestation_response(self) -> None:
        credential = """{
            "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "response": {
                "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "nfc",
                "usb"
            ]
        }"""

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.NONE
        assert verification.aaguid == "00000000-0000-0000-0000-000000000000"
        assert verification.credential_id == base64url_to_bytes(
            "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w"
        )
        assert verification.credential_public_key == base64url_to_bytes(
            "pQECAyYgASFYIEhW1CRfuNlIN6XTPKw0RbvzeaIlRMrDwwep-uq_-3WQIlgg1FZwd_RZRsqS_qgKCDvcVh7ScoKNo3w5h5fv3ihUSww"
        )
        assert verification.attestation_object == base64url_to_bytes(
            "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA"
        )
        assert verification.credential_type == PublicKeyCredentialType.PUBLIC_KEY
        assert verification.sign_count == 23
        assert verification.credential_backed_up is False
        assert verification.credential_device_type == "single_device"

    @patch("webauthn.registration.verify_registration_response.parse_attestation_object")
    def test_verifies_response_optional_user_presence(
        self,
        mock_parse_attestation_object: MagicMock,
    ) -> None:

        credential = parse_registration_credential_json(
            {
                "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
                "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                },
                "type": "public-key",
                "clientExtensionResults": {},
                "transports": ["nfc", "usb"],
            }
        )

        # Grab the actual authenticator data out of the credential above
        attestation_object = parse_attestation_object(credential.response.attestation_object)
        # Pretend this is a conditional create response
        attestation_object.auth_data.flags.up = False
        attestation_object.auth_data.flags.uv = False

        mock_parse_attestation_object.return_value = attestation_object

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
            # Be okay with up:False
            require_user_presence=False,
            require_user_verification=False,
        )

        assert verification.fmt == AttestationFormat.NONE

    def test_raises_exception_on_unsupported_attestation_type(self) -> None:
        cred_json = {
            "id": "FsWBrFcw8yRjxV8z18Egh91o1AScNRYkIuUoY6wIlIhslDpP7eydKi1q5s9g1ugDP9mqBlPDDFPRbH6YLwHbtg",
            "rawId": "FsWBrFcw8yRjxV8z18Egh91o1AScNRYkIuUoY6wIlIhslDpP7eydKi1q5s9g1ugDP9mqBlPDDFPRbH6YLwHbtg",
            "response": {
                "attestationObject": "o2NmbXRmcGFja2VkZ2F0dFN0bXSjY2FsZyZjc2lnWEcwRQIgRpuZ6hdaLAgWgCFTIo4BGSTBAxwwqk4u3s1-JAzv_H4CIQCZnfoic34aOwlac1A09eflEtb0V1kO7yGhHOw5P5wVWmN4NWOBWQLBMIICvTCCAaWgAwIBAgIEKudiYzANBgkqhkiG9w0BAQsFADAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowbjELMAkGA1UEBhMCU0UxEjAQBgNVBAoMCVl1YmljbyBBQjEiMCAGA1UECwwZQXV0aGVudGljYXRvciBBdHRlc3RhdGlvbjEnMCUGA1UEAwweWXViaWNvIFUyRiBFRSBTZXJpYWwgNzE5ODA3MDc1MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEKgOGXmBD2Z4R_xCqJVRXhL8Jr45rHjsyFykhb1USGozZENOZ3cdovf5Ke8fj2rxi5tJGn_VnW4_6iQzKdIaeP6NsMGowIgYJKwYBBAGCxAoCBBUxLjMuNi4xLjQuMS40MTQ4Mi4xLjEwEwYLKwYBBAGC5RwCAQEEBAMCBDAwIQYLKwYBBAGC5RwBAQQEEgQQbUS6m_bsLkm5MAyP6SDLczAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQByV9A83MPhFWmEkNb4DvlbUwcjc9nmRzJjKxHc3HeK7GvVkm0H4XucVDB4jeMvTke0WHb_jFUiApvpOHh5VyMx5ydwFoKKcRs5x0_WwSWL0eTZ5WbVcHkDR9pSNcA_D_5AsUKOBcbpF5nkdVRxaQHuuIuwV4k1iK2IqtMNcU8vL6w21U261xCcWwJ6sMq4zzVO8QCKCQhsoIaWrwz828GDmPzfAjFsJiLJXuYivdHACkeJ5KHMt0mjVLpfJ2BCML7_rgbmvwL7wBW80VHfNdcKmKjkLcpEiPzwcQQhiN_qHV90t-p4iyr5xRSpurlP5zic2hlRkLKxMH2_kRjhqSn4aGF1dGhEYXRhWMRJlg3liA6MaHQ0Fw9kdmBbj-SuuaKGMseZXPO6gx2XY0UAAAAqbUS6m_bsLkm5MAyP6SDLcwBAFsWBrFcw8yRjxV8z18Egh91o1AScNRYkIuUoY6wIlIhslDpP7eydKi1q5s9g1ugDP9mqBlPDDFPRbH6YLwHbtqUBAgMmIAEhWCAq3y0RWh8nLzanBZQwTA7yAbUy9KEDAM0b3N9Elrb0VCJYIJrX7ygtpyInb5mXBE7g9YEow6xWrJ400HhL2r4q5tzV",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoicERSbWtkZHVBaS1BVTJ4Nm8tRnFxaEkzWEsybmxWbHNDU3IwNHpXa050djg0SndyTUh0RWxSSEhVV0xFRGhrckVhUThCMWxCY0lIX1ZTUnFwX1JBQXciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": ["nfc", "usb"],
        }

        # Take the otherwise legitimate credential and mangle its attestationObject's
        # "fmt" to something it could never actually be
        parsed_atte_obj: dict = parse_cbor(
            base64url_to_bytes(cred_json["response"]["attestationObject"])  # type: ignore
        )
        parsed_atte_obj["fmt"] = "not_real_fmt"
        cred_json["response"]["attestationObject"] = bytes_to_base64url(  # type: ignore
            encode_cbor(parsed_atte_obj)
        )

        credential = json.dumps(cred_json)
        challenge = base64url_to_bytes(
            "pDRmkdduAi-AU2x6o-FqqhI3XK2nlVlsCSr04zWkNtv84JwrMHtElRHHUWLEDhkrEaQ8B1lBcIH_VSRqp_RAAw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        with self.assertRaises(InvalidRegistrationResponse):
            verify_registration_response(
                credential=credential,
                expected_challenge=challenge,
                expected_origin=expected_origin,
                expected_rp_id=rp_id,
            )

    def test_supports_multiple_expected_origins(self) -> None:
        credential = """{
            "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "response": {
                "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "nfc",
                "usb"
            ]
        }"""

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = ["https://foo.bar", "http://localhost:5000"]

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.credential_id == base64url_to_bytes(
            "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w"
        )

    def test_raises_when_root_cert_invalid_for_response(self) -> None:
        # "packed"
        credential = """{
            "id": "syGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k4w",
            "rawId": "syGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k4w",
            "response": {
                "attestationObject": "o2NmbXRmcGFja2VkZ2F0dFN0bXSjY2FsZyZjc2lnWEcwRQIhAOfrFlQpbavT6dJeTDJSCDzYSYPjBDHli2-syT2c1IiKAiAx5gQ2z5cHjdQX-jEHTb7JcjfQoVSW8fXszF5ihSgeOGN4NWOBWQLBMIICvTCCAaWgAwIBAgIEKudiYzANBgkqhkiG9w0BAQsFADAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowbjELMAkGA1UEBhMCU0UxEjAQBgNVBAoMCVl1YmljbyBBQjEiMCAGA1UECwwZQXV0aGVudGljYXRvciBBdHRlc3RhdGlvbjEnMCUGA1UEAwweWXViaWNvIFUyRiBFRSBTZXJpYWwgNzE5ODA3MDc1MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEKgOGXmBD2Z4R_xCqJVRXhL8Jr45rHjsyFykhb1USGozZENOZ3cdovf5Ke8fj2rxi5tJGn_VnW4_6iQzKdIaeP6NsMGowIgYJKwYBBAGCxAoCBBUxLjMuNi4xLjQuMS40MTQ4Mi4xLjEwEwYLKwYBBAGC5RwCAQEEBAMCBDAwIQYLKwYBBAGC5RwBAQQEEgQQbUS6m_bsLkm5MAyP6SDLczAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQByV9A83MPhFWmEkNb4DvlbUwcjc9nmRzJjKxHc3HeK7GvVkm0H4XucVDB4jeMvTke0WHb_jFUiApvpOHh5VyMx5ydwFoKKcRs5x0_WwSWL0eTZ5WbVcHkDR9pSNcA_D_5AsUKOBcbpF5nkdVRxaQHuuIuwV4k1iK2IqtMNcU8vL6w21U261xCcWwJ6sMq4zzVO8QCKCQhsoIaWrwz828GDmPzfAjFsJiLJXuYivdHACkeJ5KHMt0mjVLpfJ2BCML7_rgbmvwL7wBW80VHfNdcKmKjkLcpEiPzwcQQhiN_qHV90t-p4iyr5xRSpurlP5zic2hlRkLKxMH2_kRjhqSn4aGF1dGhEYXRhWMRJlg3liA6MaHQ0Fw9kdmBbj-SuuaKGMseZXPO6gx2XY0UAAAA0bUS6m_bsLkm5MAyP6SDLcwBAsyGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k46UBAgMmIAEhWCBAX_i3O3DvBnkGq_uLNk_PeAX5WwO_MIxBp0mhX6Lw7yJYIOW-1-Fch829McWvRUYAHTWZTx5IycKSGECL1UzUaK_8",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiOExCQ2lPWTNxMWNCWkhGQVd0UzRBWlpDaHpHcGh5NjdsSzdJNzB6S2k0eUM3cGdyUTJQY2g3bkFqTGsxd3E5Z3Jlc2hJQXNXMkFqaWJoWGpqSTBUbVEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "nfc",
                "usb"
            ]
        }"""
        challenge = base64url_to_bytes(
            "8LBCiOY3q1cBZHFAWtS4AZZChzGphy67lK7I70zKi4yC7pgrQ2Pch7nAjLk1wq9greshIAsW2AjibhXjjI0TmQ"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        with self.assertRaises(InvalidRegistrationResponse):
            verify_registration_response(
                credential=credential,
                expected_challenge=challenge,
                expected_origin=expected_origin,
                expected_rp_id=rp_id,
                pem_root_certs_bytes_by_fmt={
                    # This root cert is actually for android-safetynet
                    AttestationFormat.PACKED: [globalsign_r2]
                },
            )

    def test_verifies_registration_over_cable(self) -> None:
        credential = """{
            "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "response": {
                "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "cable"
            ]
        }"""

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.NONE

    def test_supports_already_parsed_credential(self) -> None:
        parsed_credential = parse_registration_credential_json(
            """{
                "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
                "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
                },
                "type": "public-key",
                "clientExtensionResults": {},
                "transports": [
                    "cable"
                ]
            }"""
        )

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=parsed_credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.NONE

    def test_supports_dict_credential(self) -> None:
        credential = {
            "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "response": {
                "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAFwAAAAAAAAAAAAAAAAAAAAAAQPctcQPE5oNRRJk_nO_371mf7qE7qIodzr0eOf6ACvnMB1oQG165dqutoi1U44shGezu5_gkTjmOPeJO0N8a7P-lAQIDJiABIVggSFbUJF-42Ug3pdM8rDRFu_N5oiVEysPDB6n66r_7dZAiWCDUVnB39FlGypL-qAoIO9xWHtJygo2jfDmHl-_eKFRLDA",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": ["cable"],
        }

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.NONE

    def test_raises_useful_error_on_bad_attestation_object(self) -> None:
        credential = {
            "id": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "rawId": "9y1xA8Tmg1FEmT-c7_fvWZ_uoTuoih3OvR45_oAK-cwHWhAbXrl2q62iLVTjiyEZ7O7n-CROOY494k7Q3xrs_w",
            "response": {
                "attestationObject": "",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiVHdON240V1R5R0tMYzRaWS1xR3NGcUtuSE00bmdscXN5VjBJQ0psTjJUTzlYaVJ5RnRya2FEd1V2c3FsLWdrTEpYUDZmbkYxTWxyWjUzTW00UjdDdnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": ["cable"],
        }

        challenge = base64url_to_bytes(
            "TwN7n4WTyGKLc4ZY-qGsFqKnHM4nglqsyV0ICJlN2TO9XiRyFtrkaDwUvsql-gkLJXP6fnF1MlrZ53Mm4R7Cvw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        with self.assertRaises(InvalidCBORData):
            verify_registration_response(
                credential=credential,
                expected_challenge=challenge,
                expected_origin=expected_origin,
                expected_rp_id=rp_id,
            )



from unittest import TestCase
from unittest.mock import MagicMock, patch

from webauthn.helpers import verify_safetynet_timestamp


class TestVerifySafetyNetTimestamp(TestCase):
    mock_time: MagicMock
    # time.time() returns time in microseconds
    mock_now = 1636589648

    def setUp(self) -> None:
        super().setUp()
        time_patch = patch("time.time")
        self.mock_time = time_patch.start()
        self.mock_time.return_value = self.mock_now

    def tearDown(self) -> None:
        super().tearDown()
        patch.stopall()

    def test_does_not_raise_on_timestamp_slightly_in_future(self):
        # Put timestamp just a bit in the future
        timestamp_ms = (self.mock_now * 1000) + 600
        verify_safetynet_timestamp(timestamp_ms)

        assert True

    def test_does_not_raise_on_timestamp_slightly_in_past(self):
        # Put timestamp just a bit in the past
        timestamp_ms = (self.mock_now * 1000) - 600
        verify_safetynet_timestamp(timestamp_ms)

        assert True

    def test_raises_on_timestamp_too_far_in_future(self):
        # Put timestamp 20 seconds in the future
        timestamp_ms = (self.mock_now * 1000) + 20000
        self.assertRaisesRegex(
            ValueError,
            "was later than",
            lambda: verify_safetynet_timestamp(timestamp_ms),
        )

    def test_raises_on_timestamp_too_far_in_past(self):
        # Put timestamp 20 seconds in the past
        timestamp_ms = (self.mock_now * 1000) - 20000
        self.assertRaisesRegex(
            ValueError,
            "expired",
            lambda: verify_safetynet_timestamp(timestamp_ms),
        )

    def test_does_not_raise_on_last_possible_millisecond(self):
        # Timestamp is verified at the exact last millisecond
        timestamp_ms = (self.mock_now * 1000) + 10000
        verify_safetynet_timestamp(timestamp_ms)

        assert True



import json
from unittest import TestCase

from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.options_to_json import options_to_json
from webauthn.helpers.structs import (
    AttestationConveyancePreference,
    AuthenticatorAttachment,
    AuthenticatorSelectionCriteria,
    AuthenticatorTransport,
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialHint,
    ResidentKeyRequirement,
    UserVerificationRequirement,
)
from webauthn import generate_registration_options, generate_authentication_options


class TestWebAuthnOptionsToJSON(TestCase):
    maxDiff = None

    def test_converts_registration_options_to_JSON(self) -> None:
        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_id=bytes([1, 2, 3, 4]),
            user_name="lee",
            user_display_name="Lee",
            attestation=AttestationConveyancePreference.DIRECT,
            authenticator_selection=AuthenticatorSelectionCriteria(
                authenticator_attachment=AuthenticatorAttachment.PLATFORM,
                resident_key=ResidentKeyRequirement.REQUIRED,
            ),
            challenge=b"1234567890",
            exclude_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            supported_pub_key_algs=[COSEAlgorithmIdentifier.ECDSA_SHA_512],
            timeout=120000,
            hints=[
                PublicKeyCredentialHint.SECURITY_KEY,
                PublicKeyCredentialHint.CLIENT_DEVICE,
                PublicKeyCredentialHint.HYBRID,
            ],
        )

        output = options_to_json(options)

        self.assertEqual(
            json.loads(output),
            {
                "rp": {"name": "Example Co", "id": "example.com"},
                "user": {
                    "id": "AQIDBA",
                    "name": "lee",
                    "displayName": "Lee",
                },
                "challenge": "MTIzNDU2Nzg5MA",
                "pubKeyCredParams": [{"type": "public-key", "alg": -36}],
                "timeout": 120000,
                "excludeCredentials": [{"type": "public-key", "id": "MTIzNDU2Nzg5MA"}],
                "authenticatorSelection": {
                    "authenticatorAttachment": "platform",
                    "residentKey": "required",
                    "requireResidentKey": True,
                    "userVerification": "preferred",
                },
                "attestation": "direct",
                "hints": ["security-key", "client-device", "hybrid"],
            },
        )

    def test_includes_optional_value_when_set(self) -> None:
        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_name="lee",
            exclude_credentials=[
                PublicKeyCredentialDescriptor(
                    id=b"1234567890",
                    transports=[AuthenticatorTransport.USB],
                )
            ],
        )

        output = options_to_json(options)

        self.assertEqual(
            json.loads(output)["excludeCredentials"],
            [
                {
                    "id": "MTIzNDU2Nzg5MA",
                    "transports": ["usb"],
                    "type": "public-key",
                }
            ],
        )

    def test_converts_authentication_options_to_JSON(self) -> None:
        options = generate_authentication_options(
            rp_id="example.com",
            challenge=b"1234567890",
            allow_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            timeout=120000,
            user_verification=UserVerificationRequirement.DISCOURAGED,
        )

        output = options_to_json(options)

        self.assertEqual(
            json.loads(output),
            {
                "rpId": "example.com",
                "challenge": "MTIzNDU2Nzg5MA",
                "allowCredentials": [{"type": "public-key", "id": "MTIzNDU2Nzg5MA"}],
                "timeout": 120000,
                "userVerification": "discouraged",
            },
        )

    def test_raises_on_bad_input(self) -> None:
        class FooClass:
            pass

        with self.assertRaisesRegex(TypeError, "not instance"):
            options_to_json(FooClass())  # type: ignore



from typing import Union


def byteslike_to_bytes(val: Union[bytes, memoryview]) -> bytes:
    """
    Massage bytes subclasses into bytes for ease of concatenation, comparison, etc...
    """
    if isinstance(val, memoryview):
        val = val.tobytes()

    return bytes(val)



import json
from json.decoder import JSONDecodeError
from typing import Union, Optional, List


from .structs import (
    PublicKeyCredentialCreationOptions,
    PublicKeyCredentialHint,
    PublicKeyCredentialRpEntity,
    PublicKeyCredentialUserEntity,
    AttestationConveyancePreference,
    AuthenticatorSelectionCriteria,
    AuthenticatorAttachment,
    ResidentKeyRequirement,
    UserVerificationRequirement,
    PublicKeyCredentialParameters,
    PublicKeyCredentialDescriptor,
    AuthenticatorTransport,
)
from .cose import COSEAlgorithmIdentifier
from .exceptions import InvalidJSONStructure, InvalidRegistrationOptions
from .base64url_to_bytes import base64url_to_bytes


def parse_registration_options_json(
    json_val: Union[str, dict]
) -> PublicKeyCredentialCreationOptions:
    """
    Parse a JSON form of registration options, as either stringified JSON or a plain dict, into an
    instance of `PublicKeyCredentialCreationOptions`. Typically useful in mapping output from
    `generate_registration_options()`, that's been persisted as JSON via Redis/etc... back into
    structured data.
    """
    if isinstance(json_val, str):
        try:
            json_val = json.loads(json_val)
        except JSONDecodeError:
            raise InvalidJSONStructure("Unable to decode options as JSON")

    if not isinstance(json_val, dict):
        raise InvalidJSONStructure("Options were not a JSON object")

    """
    Check rp
    """
    options_rp = json_val.get("rp")
    if not isinstance(options_rp, dict):
        raise InvalidJSONStructure("Options missing required rp")

    options_rp_id = options_rp.get("id")
    if options_rp_id is not None and not isinstance(options_rp_id, str):
        raise InvalidJSONStructure("Options rp.id present but not string")

    options_rp_name = options_rp.get("name")
    if not isinstance(options_rp_name, str):
        raise InvalidJSONStructure("Options rp missing required name")

    """
    Check user
    """
    options_user = json_val.get("user")
    if not isinstance(options_user, dict):
        raise InvalidJSONStructure("Options missing required user")

    options_user_id = options_user.get("id")
    if not isinstance(options_user_id, str):
        raise InvalidJSONStructure("Options user missing required id")

    options_user_name = options_user.get("name")
    if not isinstance(options_user_name, str):
        raise InvalidJSONStructure("Options user missing required name")

    options_user_display_name = options_user.get("displayName")
    if not isinstance(options_user_display_name, str):
        raise InvalidJSONStructure("Options user missing required displayName")

    """
    Check attestation
    """
    options_attestation = json_val.get("attestation")
    if not isinstance(options_attestation, str):
        raise InvalidJSONStructure("Options missing required attestation")

    try:
        mapped_attestation = AttestationConveyancePreference(options_attestation)
    except ValueError as exc:
        raise InvalidJSONStructure("Options attestation was invalid value") from exc

    """
    Check authenticatorSelection
    """
    options_authr_selection = json_val.get("authenticatorSelection")
    mapped_authenticator_selection: Optional[AuthenticatorSelectionCriteria] = None
    if isinstance(options_authr_selection, dict):
        options_authr_selection_attachment = options_authr_selection.get("authenticatorAttachment")
        mapped_attachment = None
        if options_authr_selection_attachment is not None:
            try:
                mapped_attachment = AuthenticatorAttachment(options_authr_selection_attachment)
            except ValueError as exc:
                raise InvalidJSONStructure(
                    "Options authenticatorSelection attachment was invalid value"
                ) from exc

        options_authr_selection_rkey = options_authr_selection.get("residentKey")
        mapped_rkey = None
        if options_authr_selection_rkey is not None:
            try:
                mapped_rkey = ResidentKeyRequirement(options_authr_selection_rkey)
            except ValueError as exc:
                raise InvalidJSONStructure(
                    "Options authenticatorSelection residentKey was invalid value"
                ) from exc

        options_authr_selection_require_rkey = options_authr_selection.get("requireResidentKey")
        mapped_require_rkey = False
        if options_authr_selection_require_rkey is not None:
            if not isinstance(options_authr_selection_require_rkey, bool):
                raise InvalidJSONStructure(
                    "Options authenticatorSelection requireResidentKey was invalid boolean"
                )

            mapped_require_rkey = options_authr_selection_require_rkey

        options_authr_selection_uv = options_authr_selection.get("userVerification")
        mapped_user_verification = UserVerificationRequirement.PREFERRED
        if options_authr_selection_uv is not None:
            try:
                mapped_user_verification = UserVerificationRequirement(options_authr_selection_uv)
            except ValueError as exc:
                raise InvalidJSONStructure(
                    "Options authenticatorSelection userVerification was invalid value"
                ) from exc

        mapped_authenticator_selection = AuthenticatorSelectionCriteria(
            authenticator_attachment=mapped_attachment,
            resident_key=mapped_rkey,
            require_resident_key=mapped_require_rkey,
            user_verification=mapped_user_verification,
        )

    """
    Check challenge is present
    """
    options_challenge = json_val.get("challenge")
    if not isinstance(options_challenge, str):
        raise InvalidJSONStructure("Options missing required challenge")

    """
    Check pubKeyCredParams
    """
    options_pub_key_cred_params = json_val.get("pubKeyCredParams")
    if not isinstance(options_pub_key_cred_params, list):
        raise InvalidJSONStructure("Options pubKeyCredParams was invalid value")

    try:
        mapped_pub_key_cred_params = [
            PublicKeyCredentialParameters(
                alg=COSEAlgorithmIdentifier(param["alg"]), type="public-key"
            )
            for param in options_pub_key_cred_params
        ]
    except ValueError as exc:
        raise InvalidJSONStructure("Options pubKeyCredParams entry had invalid alg") from exc

    """
    Check excludeCredentials
    """
    options_exclude_credentials = json_val.get("excludeCredentials")
    mapped_exclude_credentials: Optional[List[PublicKeyCredentialDescriptor]] = None
    if isinstance(options_exclude_credentials, list):
        mapped_exclude_credentials = []
        for cred in options_exclude_credentials:
            _cred_id = cred.get("id")
            if not isinstance(_cred_id, str):
                raise InvalidJSONStructure("Options excludeCredentials entry missing required id")

            _mapped = PublicKeyCredentialDescriptor(id=base64url_to_bytes(_cred_id))

            _transports = cred.get("transports")
            if _transports is not None:
                if not isinstance(_transports, list):
                    raise InvalidJSONStructure(
                        "Options excludeCredentials entry transports was not list"
                    )
                try:
                    _mapped.transports = [
                        AuthenticatorTransport(_transport) for _transport in _transports
                    ]
                except ValueError as exc:
                    raise InvalidJSONStructure(
                        "Options excludeCredentials entry transports had invalid value"
                    ) from exc

            mapped_exclude_credentials.append(_mapped)

    """
    Check timeout
    """
    options_timeout = json_val.get("timeout")
    mapped_timeout = None
    if isinstance(options_timeout, int):
        mapped_timeout = options_timeout

    """
    Check hints
    """
    options_hints = json_val.get("hints")
    mapped_hints = None
    if options_hints is not None:
        if not isinstance(options_hints, list):
            raise InvalidJSONStructure("Options hints was invalid value")

        try:
            mapped_hints = [PublicKeyCredentialHint(hint) for hint in options_hints]
        except ValueError as exc:
            raise InvalidJSONStructure("Options hints had invalid value") from exc

    try:
        registration_options = PublicKeyCredentialCreationOptions(
            rp=PublicKeyCredentialRpEntity(
                id=options_rp_id,
                name=options_rp_name,
            ),
            user=PublicKeyCredentialUserEntity(
                id=base64url_to_bytes(options_user_id),
                name=options_user_name,
                display_name=options_user_display_name,
            ),
            attestation=mapped_attestation,
            authenticator_selection=mapped_authenticator_selection,
            challenge=base64url_to_bytes(options_challenge),
            pub_key_cred_params=mapped_pub_key_cred_params,
            exclude_credentials=mapped_exclude_credentials,
            timeout=mapped_timeout,
            hints=mapped_hints,
        )
    except Exception as exc:
        raise InvalidRegistrationOptions(
            "Could not parse registration options from JSON data"
        ) from exc

    return registration_options



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.exceptions import InvalidJSONStructure, InvalidRegistrationResponse
from webauthn.helpers.structs import AuthenticatorTransport, AuthenticatorAttachment
from webauthn.helpers.parse_registration_credential_json import parse_registration_credential_json


class TestParseRegistrationCredentialJSON(TestCase):
    def test_raises_on_non_dict_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "not a JSON object"):
            parse_registration_credential_json("[0]")

    def test_raises_on_missing_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required id"):
            parse_registration_credential_json({})

    def test_raises_on_missing_raw_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required rawId"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                }
            )

    def test_raises_on_missing_response(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required response"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                }
            )

    def test_raises_on_missing_client_data_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required clientDataJSON"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {},
                }
            )

    def test_raises_on_missing_attestation_object(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required attestationObject"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "clientDataJSON": "...",
                    },
                }
            )

    def test_validates_credential_type(self) -> None:
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "...",
                    "clientDataJSON": "...",
                },
                "type": "public-key",
            }
        )

    def test_raises_on_invalid_credential_type(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "unexpected type"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "attestationObject": "...",
                        "clientDataJSON": "...",
                    },
                    "type": "not-a-public-key",
                }
            )

    def test_parses_transports(self) -> None:
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                    "transports": ["internal", "hybrid"],
                },
                "type": "public-key",
            }
        )

        self.assertEqual(
            parsed.response.transports,
            [AuthenticatorTransport.INTERNAL, AuthenticatorTransport.HYBRID],
        )

    def test_handles_missing_transports(self) -> None:
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                },
                "type": "public-key",
            }
        )

        self.assertIsNone(parsed.response.transports)

    def test_ignores_non_list_transports(self) -> None:
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                    # Pretend someone got clever on the front end
                    "transports": "usb|nfc|ble",
                },
                "type": "public-key",
            }
        )

        self.assertIsNone(parsed.response.transports)

    def test_handles_authenticator_attachment(self) -> None:
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                },
                "authenticatorAttachment": "platform",
                "type": "public-key",
            }
        )

        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)

    def test_handles_bad_authenticator_attachment(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "unexpected authenticatorAttachment"):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                        "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                    },
                    "authenticatorAttachment": "badValue",
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_raw_id(self) -> None:
        with self.assertRaisesRegex(
            InvalidRegistrationResponse, "Could not parse registration credential"
        ):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "baddd",
                    "response": {
                        "attestationObject": "...",
                        "clientDataJSON": "...",
                    },
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_attestation_object(self) -> None:
        with self.assertRaisesRegex(
            InvalidRegistrationResponse, "Could not parse registration credential"
        ):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "attestationObject": "baddd",
                        "clientDataJSON": "...",
                    },
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_client_data_json(self) -> None:
        with self.assertRaisesRegex(
            InvalidRegistrationResponse, "Could not parse registration credential"
        ):
            parse_registration_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "attestationObject": "...",
                        "clientDataJSON": "baddd",
                    },
                    "type": "public-key",
                }
            )

    def test_success_from_dict(self) -> None:
        # A bit more complex than a basic response, but it should get parsed all the same
        parsed = parse_registration_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                    "transports": ["internal"],
                    "publicKeyAlgorithm": -7,
                    "publicKey": "MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE6ZHW_UqFyItksz8MPHLrswNrcnW-yJBbLOXbQ_SzLp-K1r4e9lWKdFVRFpYKv4YkOb2JD5nV_lZ9GE4DcLBlxQ",
                    "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                },
                "type": "public-key",
                "clientExtensionResults": {"credProps": {"rk": True}},
                "authenticatorAttachment": "platform",
            }
        )

        self.assertEqual(parsed.id, "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A")
        self.assertEqual(
            parsed.raw_id,
            base64url_to_bytes("Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A"),
        )
        self.assertEqual(
            parsed.response.attestation_object,
            base64url_to_bytes(
                "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU"
            ),
        )
        self.assertEqual(
            parsed.response.client_data_json,
            base64url_to_bytes(
                "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0"
            ),
        )
        self.assertEqual(
            parsed.response.transports,
            [AuthenticatorTransport.INTERNAL],
        )
        self.assertEqual(parsed.type, "public-key")
        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)

    def test_success_from_str(self) -> None:
        # Same dict as above, just stringified
        parsed = parse_registration_credential_json(
            """{
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0",
                    "transports": ["internal"],
                    "publicKeyAlgorithm": -7,
                    "publicKey": "MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE6ZHW_UqFyItksz8MPHLrswNrcnW-yJBbLOXbQ_SzLp-K1r4e9lWKdFVRFpYKv4YkOb2JD5nV_lZ9GE4DcLBlxQ",
                    "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU"
                },
                "type": "public-key",
                "clientExtensionResults": {"credProps": {"rk": true}},
                "authenticatorAttachment": "platform"
            }"""
        )

        self.assertEqual(parsed.id, "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A")
        self.assertEqual(
            parsed.raw_id,
            base64url_to_bytes("Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A"),
        )
        self.assertEqual(
            parsed.response.attestation_object,
            base64url_to_bytes(
                "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVikdKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvBBAAAAAK3OAAI1vMYKZIsLJfHwVQMAIFH-jHEbNNz948Ezg2yqJF5T6_kME1M0b-NZf85uvdvwpQECAyYgASFYIOmR1v1KhciLZLM_DDxy67MDa3J1vsiQWyzl20P0sy6fIlggita-HvZVinRVURaWCr-GJDm9iQ-Z1f5WfRhOA3CwZcU"
            ),
        )
        self.assertEqual(
            parsed.response.client_data_json,
            base64url_to_bytes(
                "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiUENYcURTMDFJbkRTeFAwbDhCMnVDcWxoR1BseEw4VHJSeDdpbHpjczIwNHplTklvMlJ0U3RkbFVSMWhfaW5WQzVPYkNjOElNT1JSYl9jWWNJMDNNeFEiLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlLCJvdGhlcl9rZXlzX2Nhbl9iZV9hZGRlZF9oZXJlIjoiZG8gbm90IGNvbXBhcmUgY2xpZW50RGF0YUpTT04gYWdhaW5zdCBhIHRlbXBsYXRlLiBTZWUgaHR0cHM6Ly9nb28uZ2wveWFiUGV4In0"
            ),
        )
        self.assertEqual(
            parsed.response.transports,
            [AuthenticatorTransport.INTERNAL],
        )
        self.assertEqual(parsed.type, "public-key")
        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)



import json
from json.decoder import JSONDecodeError
from typing import Union

from .base64url_to_bytes import base64url_to_bytes
from .byteslike_to_bytes import byteslike_to_bytes
from .exceptions import InvalidJSONStructure
from .structs import CollectedClientData, TokenBinding


def parse_client_data_json(val: bytes) -> CollectedClientData:
    """
    Break apart `response.clientDataJSON` buffer into structured data
    """
    val = byteslike_to_bytes(val)

    try:
        json_dict = json.loads(val)
    except JSONDecodeError:
        raise InvalidJSONStructure("Unable to decode client_data_json bytes as JSON")

    # Ensure required values are present in client data
    if "type" not in json_dict:
        raise InvalidJSONStructure('client_data_json missing required property "type"')
    if "challenge" not in json_dict:
        raise InvalidJSONStructure('client_data_json missing required property "challenge"')
    if "origin" not in json_dict:
        raise InvalidJSONStructure('client_data_json missing required property "origin"')

    client_data = CollectedClientData(
        type=json_dict["type"],
        challenge=base64url_to_bytes(json_dict["challenge"]),
        origin=json_dict["origin"],
    )

    # Populate optional values if set
    if "crossOrigin" in json_dict:
        cross_origin = bool(json_dict["crossOrigin"])
        client_data.cross_origin = cross_origin

    if "tokenBinding" in json_dict:
        token_binding_dict = json_dict["tokenBinding"]

        # Some U2F devices set a string to `token_binding`, in which case ignore it
        if type(token_binding_dict) is dict:
            if "status" not in token_binding_dict:
                raise InvalidJSONStructure('token_binding missing required property "status"')

            status = token_binding_dict["status"]
            try:
                # This will raise ValidationError on an unexpected status
                token_binding = TokenBinding(status=status)

                # Handle optional values
                if "id" in token_binding_dict:
                    id = token_binding_dict["id"]
                    token_binding.id = f"{id}"

                client_data.token_binding = token_binding
            except Exception:
                # If we encounter a status we don't expect then ignore token_binding
                # completely
                pass

    return client_data



from .registration.generate_registration_options import generate_registration_options
from .registration.verify_registration_response import verify_registration_response
from .authentication.generate_authentication_options import (
    generate_authentication_options,
)
from .authentication.verify_authentication_response import (
    verify_authentication_response,
)
from .helpers import base64url_to_bytes, options_to_json

__version__ = "2.5.3"

__all__ = [
    "generate_registration_options",
    "verify_registration_response",
    "generate_authentication_options",
    "verify_authentication_response",
    "base64url_to_bytes",
    "options_to_json",
]



from unittest import TestCase

from webauthn.helpers import parse_backup_flags
from webauthn.helpers.structs import AuthenticatorDataFlags
from webauthn.helpers.exceptions import InvalidBackupFlags


class TestParseBackupFlags(TestCase):
    flags: AuthenticatorDataFlags

    def setUp(self) -> None:
        self.flags = AuthenticatorDataFlags(
            up=True,
            uv=False,
            be=False,
            bs=False,
            at=False,
            ed=False,
        )

    def test_returns_single_device_not_backed_up(self) -> None:
        self.flags.be = False
        self.flags.bs = False

        parsed = parse_backup_flags(self.flags)

        self.assertEqual(parsed.credential_device_type, "single_device")
        self.assertEqual(parsed.credential_backed_up, False)

    def test_returns_multi_device_not_backed_up(self) -> None:
        self.flags.be = True
        self.flags.bs = False

        parsed = parse_backup_flags(self.flags)

        self.assertEqual(parsed.credential_device_type, "multi_device")
        self.assertEqual(parsed.credential_backed_up, False)

    def test_returns_multi_device_backed_up(self) -> None:
        self.flags.be = True
        self.flags.bs = True

        parsed = parse_backup_flags(self.flags)

        self.assertEqual(parsed.credential_device_type, "multi_device")
        self.assertEqual(parsed.credential_backed_up, True)

    def test_raises_on_invalid_backup_state_flags(self) -> None:
        self.flags.be = False
        self.flags.bs = True

        with self.assertRaisesRegex(
            InvalidBackupFlags,
            "impossible",
        ):
            parse_backup_flags(self.flags)



from dataclasses import dataclass
import hashlib
from typing import List, Union

from cryptography.exceptions import InvalidSignature

from webauthn.helpers import (
    bytes_to_base64url,
    byteslike_to_bytes,
    decode_credential_public_key,
    decoded_public_key_to_cryptography,
    parse_authenticator_data,
    parse_backup_flags,
    parse_client_data_json,
    parse_authentication_credential_json,
    verify_signature,
)
from webauthn.helpers.exceptions import InvalidAuthenticationResponse
from webauthn.helpers.structs import (
    AuthenticationCredential,
    ClientDataType,
    CredentialDeviceType,
    PublicKeyCredentialType,
    TokenBindingStatus,
)


@dataclass
class VerifiedAuthentication:
    """
    Information about a verified authentication of which an RP can make use
    """

    credential_id: bytes
    new_sign_count: int
    credential_device_type: CredentialDeviceType
    credential_backed_up: bool
    user_verified: bool


expected_token_binding_statuses = [
    TokenBindingStatus.SUPPORTED,
    TokenBindingStatus.PRESENT,
]


def verify_authentication_response(
    *,
    credential: Union[str, dict, AuthenticationCredential],
    expected_challenge: bytes,
    expected_rp_id: str,
    expected_origin: Union[str, List[str]],
    credential_public_key: bytes,
    credential_current_sign_count: int,
    require_user_verification: bool = False,
) -> VerifiedAuthentication:
    """Verify a response from navigator.credentials.get()

    Args:
        - `credential`: The value returned from `navigator.credentials.get()`. Can be either a
          stringified JSON object, a plain dict, or an instance of RegistrationCredential
        - `expected_challenge`: The challenge passed to the authenticator within the preceding
          authentication options.
        - `expected_rp_id`: The Relying Party's unique identifier as specified in the preceding
          authentication options.
        - `expected_origin`: The domain, with HTTP protocol (e.g. "https://domain.here"), on which
          the authentication ceremony should have occurred.
        - `credential_public_key`: The public key for the credential's ID as provided in a
          preceding authenticator registration ceremony.
        - `credential_current_sign_count`: The current known number of times the authenticator was
          used.
        - (optional) `require_user_verification`: Whether or not to require that the authenticator
          verified the user.

    Returns:
        Information about the authenticator

    Raises:
        `helpers.exceptions.InvalidAuthenticationResponse` if the response cannot be verified
    """
    if isinstance(credential, str) or isinstance(credential, dict):
        credential = parse_authentication_credential_json(credential)

    # FIDO-specific check
    if bytes_to_base64url(credential.raw_id) != credential.id:
        raise InvalidAuthenticationResponse("id and raw_id were not equivalent")

    # FIDO-specific check
    if credential.type != PublicKeyCredentialType.PUBLIC_KEY:
        raise InvalidAuthenticationResponse(
            f'Unexpected credential type "{credential.type}", expected "public-key"'
        )

    response = credential.response

    client_data_bytes = byteslike_to_bytes(response.client_data_json)
    authenticator_data_bytes = byteslike_to_bytes(response.authenticator_data)
    signature_bytes = byteslike_to_bytes(response.signature)

    client_data = parse_client_data_json(client_data_bytes)

    if client_data.type != ClientDataType.WEBAUTHN_GET:
        raise InvalidAuthenticationResponse(
            f'Unexpected client data type "{client_data.type}", expected "{ClientDataType.WEBAUTHN_GET}"'
        )

    if expected_challenge != client_data.challenge:
        raise InvalidAuthenticationResponse("Client data challenge was not expected challenge")

    if isinstance(expected_origin, str):
        if expected_origin != client_data.origin:
            raise InvalidAuthenticationResponse(
                f'Unexpected client data origin "{client_data.origin}", expected "{expected_origin}"'
            )
    else:
        try:
            expected_origin.index(client_data.origin)
        except ValueError:
            raise InvalidAuthenticationResponse(
                f'Unexpected client data origin "{client_data.origin}", expected one of {expected_origin}'
            )

    if client_data.token_binding:
        status = client_data.token_binding.status
        if status not in expected_token_binding_statuses:
            raise InvalidAuthenticationResponse(
                f'Unexpected token_binding status of "{status}", expected one of "{",".join(expected_token_binding_statuses)}"'
            )

    auth_data = parse_authenticator_data(authenticator_data_bytes)

    # Generate a hash of the expected RP ID for comparison
    expected_rp_id_hash = hashlib.sha256()
    expected_rp_id_hash.update(expected_rp_id.encode("utf-8"))
    expected_rp_id_hash_bytes = expected_rp_id_hash.digest()

    if auth_data.rp_id_hash != expected_rp_id_hash_bytes:
        raise InvalidAuthenticationResponse("Unexpected RP ID hash")

    if not auth_data.flags.up:
        raise InvalidAuthenticationResponse("User was not present during authentication")

    if require_user_verification and not auth_data.flags.uv:
        raise InvalidAuthenticationResponse(
            "User verification is required but user was not verified during authentication"
        )

    if (
        auth_data.sign_count > 0 or credential_current_sign_count > 0
    ) and auth_data.sign_count <= credential_current_sign_count:
        # Require the sign count to have been incremented over what was reported by the
        # authenticator the last time this credential was used, otherwise this might be
        # a replay attack
        raise InvalidAuthenticationResponse(
            f"Response sign count of {auth_data.sign_count} was not greater than current count of {credential_current_sign_count}"
        )

    client_data_hash = hashlib.sha256()
    client_data_hash.update(client_data_bytes)
    client_data_hash_bytes = client_data_hash.digest()

    signature_base = authenticator_data_bytes + client_data_hash_bytes

    try:
        decoded_public_key = decode_credential_public_key(credential_public_key)
        crypto_public_key = decoded_public_key_to_cryptography(decoded_public_key)

        verify_signature(
            public_key=crypto_public_key,
            signature_alg=decoded_public_key.alg,
            signature=signature_bytes,
            data=signature_base,
        )
    except InvalidSignature:
        raise InvalidAuthenticationResponse("Could not verify authentication signature")

    parsed_backup_flags = parse_backup_flags(auth_data.flags)

    return VerifiedAuthentication(
        credential_id=credential.raw_id,
        new_sign_count=auth_data.sign_count,
        credential_device_type=parsed_backup_flags.credential_device_type,
        credential_backed_up=parsed_backup_flags.credential_backed_up,
        user_verified=auth_data.flags.uv,
    )



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.exceptions import InvalidJSONStructure, InvalidAuthenticationResponse
from webauthn.helpers.structs import AuthenticatorTransport, AuthenticatorAttachment
from webauthn.helpers.parse_authentication_credential_json import (
    parse_authentication_credential_json,
)


class TestParseClientDataJSON(TestCase):
    def test_raises_on_non_dict_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "not a JSON object"):
            parse_authentication_credential_json("[0]")

    def test_raises_on_missing_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required id"):
            parse_authentication_credential_json({})

    def test_raises_on_missing_raw_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required rawId"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                }
            )

    def test_raises_on_missing_response(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required response"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                }
            )

    def test_raises_on_missing_client_data_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required clientDataJSON"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {},
                }
            )

    def test_raises_on_missing_authenticator_data(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required authenticatorData"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "clientDataJSON": "...",
                    },
                }
            )

    def test_raises_on_missing_signature(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required signature"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "...",
                        "clientDataJSON": "...",
                    },
                }
            )

    def test_validates_credential_type(self) -> None:
        parsed = parse_authentication_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "...",
                    "clientDataJSON": "...",
                    "signature": "...",
                },
                "type": "public-key",
            }
        )

    def test_raises_on_invalid_credential_type(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "unexpected type"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "...",
                        "clientDataJSON": "...",
                        "signature": "...",
                    },
                    "type": "not-a-public-key",
                }
            )

    def test_handles_authenticator_attachment(self) -> None:
        parsed = parse_authentication_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ",
                    "signature": "MEQCIBYTvMC-3pw88hoiYwdmPCHmPdz__tuhkFrfq-E03NvSAiBzelRNe6FCgsYL6_x6xmUlWM_ULmxRi6cX5iPZPiDrxA",
                },
                "authenticatorAttachment": "platform",
                "type": "public-key",
            }
        )

        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)

    def test_handles_bad_authenticator_attachment(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "unexpected authenticatorAttachment"):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA",
                        "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ",
                        "signature": "MEQCIBYTvMC-3pw88hoiYwdmPCHmPdz__tuhkFrfq-E03NvSAiBzelRNe6FCgsYL6_x6xmUlWM_ULmxRi6cX5iPZPiDrxA",
                    },
                    "authenticatorAttachment": "badValue",
                    "type": "public-key",
                }
            )

    def test_handles_user_handle(self) -> None:
        parsed = parse_authentication_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "...",
                    "clientDataJSON": "...",
                    "signature": "...",
                    "userHandle": "bW1pbGxlcg",
                },
                "type": "public-key",
            }
        )

        self.assertEqual(parsed.response.user_handle, base64url_to_bytes("bW1pbGxlcg"))

    def test_handles_missing_user_handle(self) -> None:
        parsed = parse_authentication_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "...",
                    "clientDataJSON": "...",
                    "signature": "...",
                },
                "type": "public-key",
            }
        )

        self.assertIsNone(parsed.response.user_handle)

    def test_raises_on_non_base64url_raw_id(self) -> None:
        with self.assertRaisesRegex(
            InvalidAuthenticationResponse, "Could not parse authentication credential"
        ):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "baddd",
                    "response": {
                        "authenticatorData": "...",
                        "clientDataJSON": "...",
                        "signature": "...",
                    },
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_authenticator_data(self) -> None:
        with self.assertRaisesRegex(
            InvalidAuthenticationResponse, "Could not parse authentication credential"
        ):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "baddd",
                        "clientDataJSON": "...",
                        "signature": "...",
                    },
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_client_data_json(self) -> None:
        with self.assertRaisesRegex(
            InvalidAuthenticationResponse, "Could not parse authentication credential"
        ):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "...",
                        "clientDataJSON": "baddd",
                        "signature": "...",
                    },
                    "type": "public-key",
                }
            )

    def test_raises_on_non_base64url_signature(self) -> None:
        with self.assertRaisesRegex(
            InvalidAuthenticationResponse, "Could not parse authentication credential"
        ):
            parse_authentication_credential_json(
                {
                    "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                    "response": {
                        "authenticatorData": "...",
                        "clientDataJSON": "...",
                        "signature": "baddd",
                    },
                    "type": "public-key",
                }
            )

    def test_success_from_dict(self) -> None:
        # A bit more complex than a basic response, but it should get parsed all the same
        parsed = parse_authentication_credential_json(
            {
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ",
                    "signature": "MEQCIBYTvMC-3pw88hoiYwdmPCHmPdz__tuhkFrfq-E03NvSAiBzelRNe6FCgsYL6_x6xmUlWM_ULmxRi6cX5iPZPiDrxA",
                    "userHandle": "bW1pbGxlcg",
                },
                "type": "public-key",
                "clientExtensionResults": {},
                "authenticatorAttachment": "platform",
            }
        )

        self.assertEqual(parsed.id, "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A")
        self.assertEqual(
            parsed.raw_id,
            base64url_to_bytes("Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A"),
        )
        self.assertEqual(
            parsed.response.authenticator_data,
            base64url_to_bytes("dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA"),
        )
        self.assertEqual(
            parsed.response.client_data_json,
            base64url_to_bytes(
                "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ"
            ),
        )
        self.assertEqual(parsed.response.user_handle, base64url_to_bytes("bW1pbGxlcg"))
        self.assertEqual(parsed.type, "public-key")
        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)

    def test_success_from_str(self) -> None:
        # Same dict as above, just stringified
        parsed = parse_authentication_credential_json(
            """{
                "id": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "rawId": "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A",
                "response": {
                    "authenticatorData": "dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ",
                    "signature": "MEQCIBYTvMC-3pw88hoiYwdmPCHmPdz__tuhkFrfq-E03NvSAiBzelRNe6FCgsYL6_x6xmUlWM_ULmxRi6cX5iPZPiDrxA",
                    "userHandle": "bW1pbGxlcg"
                },
                "type": "public-key",
                "clientExtensionResults": {},
                "authenticatorAttachment": "platform"
            }"""
        )

        self.assertEqual(parsed.id, "Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A")
        self.assertEqual(
            parsed.raw_id,
            base64url_to_bytes("Uf6McRs03P3jwTODbKokXlPr-QwTUzRv41l_zm692_A"),
        )
        self.assertEqual(
            parsed.response.authenticator_data,
            base64url_to_bytes("dKbqkhPJnC90siSSsyDPQCYqlMGpUKA5fyklC2CEHvABAAAAAA"),
        )
        self.assertEqual(
            parsed.response.client_data_json,
            base64url_to_bytes(
                "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiSjlyUFpWWnFWODlUSW53bzV3cU11R3dlZjdET0pZRi1OVHlMQnhHV2pjZi16amFzOFRTUTlMbXI3em4wSmpkMTQyMU1sV0ItS2JYdEs5RW5sN19JM3ciLCJvcmlnaW4iOiJodHRwczovL3dlYmF1dGhuLmlvIiwiY3Jvc3NPcmlnaW4iOmZhbHNlfQ"
            ),
        )
        self.assertEqual(parsed.response.user_handle, base64url_to_bytes("bW1pbGxlcg"))
        self.assertEqual(parsed.type, "public-key")
        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)



import json
from json.decoder import JSONDecodeError
from typing import Union, Optional, List

from .base64url_to_bytes import base64url_to_bytes
from .exceptions import InvalidRegistrationResponse, InvalidJSONStructure
from .structs import (
    AuthenticatorAttachment,
    AuthenticatorAttestationResponse,
    AuthenticatorTransport,
    PublicKeyCredentialType,
    RegistrationCredential,
)


def parse_registration_credential_json(json_val: Union[str, dict]) -> RegistrationCredential:
    """
    Parse a JSON form of a registration credential, as either a stringified JSON object or a
    plain dict, into an instance of RegistrationCredential
    """
    if isinstance(json_val, str):
        try:
            json_val = json.loads(json_val)
        except JSONDecodeError:
            raise InvalidJSONStructure("Unable to decode credential as JSON")

    if not isinstance(json_val, dict):
        raise InvalidJSONStructure("Credential was not a JSON object")

    cred_id = json_val.get("id")
    if not isinstance(cred_id, str):
        raise InvalidJSONStructure("Credential missing required id")

    cred_raw_id = json_val.get("rawId")
    if not isinstance(cred_raw_id, str):
        raise InvalidJSONStructure("Credential missing required rawId")

    cred_response = json_val.get("response")
    if not isinstance(cred_response, dict):
        raise InvalidJSONStructure("Credential missing required response")

    response_client_data_json = cred_response.get("clientDataJSON")
    if not isinstance(response_client_data_json, str):
        raise InvalidJSONStructure("Credential response missing required clientDataJSON")

    response_attestation_object = cred_response.get("attestationObject")
    if not isinstance(response_attestation_object, str):
        raise InvalidJSONStructure("Credential response missing required attestationObject")

    cred_type = json_val.get("type")
    try:
        # Simply try to get the single matching Enum. We'll set the literal value below assuming
        # the code can get past here (this is basically a mypy optimization)
        PublicKeyCredentialType(cred_type)
    except ValueError as cred_type_exc:
        raise InvalidJSONStructure("Credential had unexpected type") from cred_type_exc

    transports: Optional[List[AuthenticatorTransport]] = None
    response_transports = cred_response.get("transports")
    if isinstance(response_transports, list):
        transports = []
        for val in response_transports:
            try:
                transport_enum = AuthenticatorTransport(val)
                transports.append(transport_enum)
            except ValueError:
                pass

    cred_authenticator_attachment = json_val.get("authenticatorAttachment")
    if isinstance(cred_authenticator_attachment, str):
        try:
            cred_authenticator_attachment = AuthenticatorAttachment(cred_authenticator_attachment)
        except ValueError as cred_attachment_exc:
            raise InvalidJSONStructure(
                "Credential had unexpected authenticatorAttachment"
            ) from cred_attachment_exc
    else:
        cred_authenticator_attachment = None

    try:
        registration_credential = RegistrationCredential(
            id=cred_id,
            raw_id=base64url_to_bytes(cred_raw_id),
            response=AuthenticatorAttestationResponse(
                client_data_json=base64url_to_bytes(response_client_data_json),
                attestation_object=base64url_to_bytes(response_attestation_object),
                transports=transports,
            ),
            authenticator_attachment=cred_authenticator_attachment,
            type=PublicKeyCredentialType.PUBLIC_KEY,
        )
    except Exception as exc:
        raise InvalidRegistrationResponse(
            "Could not parse registration credential from JSON data"
        ) from exc

    return registration_credential



from unittest import TestCase

from webauthn import verify_authentication_response, base64url_to_bytes
from webauthn.helpers.structs import (
    AuthenticationCredential,
    AuthenticatorAssertionResponse,
)


class TestWebAuthnBytesSubclassSupport(TestCase):
    def test_handles_bytes_subclasses(self) -> None:
        """
        Ensure the library can support being used in projects that might work with values that are
        subclasses of `bytes`. Let's embrace Python's duck-typing, not shy away from it
        """

        class CustomBytes(bytes):
            def __new__(cls, data: str):
                data_bytes = base64url_to_bytes(data)
                self = bytes.__new__(cls, memoryview(data_bytes).tobytes())
                return self

        verification = verify_authentication_response(
            credential=AuthenticationCredential(
                id="fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA",
                raw_id=CustomBytes(
                    "fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA"
                ),
                response=AuthenticatorAssertionResponse(
                    authenticator_data=CustomBytes(
                        "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MBAAAABw"
                    ),
                    client_data_json=CustomBytes(
                        "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiZVo0ZWVBM080ank1Rkl6cURhU0o2SkROR3UwYkJjNXpJMURqUV9rTHNvMVdOcWtHNms1bUNZZjFkdFFoVlVpQldaV2xaa3pSNU1GZWVXQ3BKUlVOWHciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
                    ),
                    signature=CustomBytes(
                        "RRWV8mYDRvK7YdQgdtZD4pJ2dh1D_IWZ_D6jsZo6FHJBoenbj0CVT5nA20vUzlRhN4R6dOEUHmUwP1F8eRBhBg"
                    ),
                ),
            ),
            expected_challenge=CustomBytes(
                "eZ4eeA3O4jy5FIzqDaSJ6JDNGu0bBc5zI1DjQ_kLso1WNqkG6k5mCYf1dtQhVUiBWZWlZkzR5MFeeWCpJRUNXw"
            ),
            expected_rp_id="localhost",
            expected_origin="http://localhost:5000",
            credential_public_key=CustomBytes(
                "pAEBAycgBiFYIMz6_SUFLiDid2Yhlq0YboyJ-CDrIrNpkPUGmJp4D3Dp"
            ),
            credential_current_sign_count=3,
        )

        assert verification.new_sign_count == 7

    def test_handles_memoryviews(self) -> None:
        """
        Ensure support for libraries that leverage memoryviews
        """

        def base64url_to_memoryview(data: str) -> memoryview:
            data_bytes = base64url_to_bytes(data)
            return memoryview(data_bytes)

        verification = verify_authentication_response(
            credential=AuthenticationCredential(
                id="fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA",
                raw_id=base64url_to_memoryview(
                    "fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA"
                ),
                response=AuthenticatorAssertionResponse(
                    authenticator_data=base64url_to_memoryview(
                        "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MBAAAABw"
                    ),
                    client_data_json=base64url_to_memoryview(
                        "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiZVo0ZWVBM080ank1Rkl6cURhU0o2SkROR3UwYkJjNXpJMURqUV9rTHNvMVdOcWtHNms1bUNZZjFkdFFoVlVpQldaV2xaa3pSNU1GZWVXQ3BKUlVOWHciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
                    ),
                    signature=base64url_to_memoryview(
                        "RRWV8mYDRvK7YdQgdtZD4pJ2dh1D_IWZ_D6jsZo6FHJBoenbj0CVT5nA20vUzlRhN4R6dOEUHmUwP1F8eRBhBg"
                    ),
                ),
            ),
            expected_challenge=base64url_to_memoryview(
                "eZ4eeA3O4jy5FIzqDaSJ6JDNGu0bBc5zI1DjQ_kLso1WNqkG6k5mCYf1dtQhVUiBWZWlZkzR5MFeeWCpJRUNXw"
            ),
            expected_rp_id="localhost",
            expected_origin="http://localhost:5000",
            credential_public_key=base64url_to_memoryview(
                "pAEBAycgBiFYIMz6_SUFLiDid2Yhlq0YboyJ-CDrIrNpkPUGmJp4D3Dp"
            ),
            credential_current_sign_count=3,
        )

        assert verification.new_sign_count == 7



from dataclasses import dataclass
from enum import Enum
from typing import Mapping, Union

from ..cose import COSECRV, COSEAlgorithmIdentifier
from ..exceptions import InvalidTPMPubAreaStructure

################
#
# A whole lotta domain knowledge is captured here, with hazy connections to source
# documents. Good places to start searching for more info on these values are the
# following Trusted Computing Group TPM Library docs linked in the WebAuthn API:
#
# - https://www.trustedcomputinggroup.org/wp-content/uploads/TPM-Rev-2.0-Part-1-Architecture-01.38.pdf
# - https://www.trustedcomputinggroup.org/wp-content/uploads/TPM-Rev-2.0-Part-2-Structures-01.38.pdf
# - https://www.trustedcomputinggroup.org/wp-content/uploads/TPM-Rev-2.0-Part-3-Commands-01.38.pdf
#
################


class TPM_ST(str, Enum):
    RSP_COMMAND = "TPM_ST_RSP_COMMAND"
    NULL = "TPM_ST_NULL"
    NO_SESSIONS = "TPM_ST_NO_SESSIONS"
    SESSIONS = "TPM_ST_SESSIONS"
    ATTEST_NV = "TPM_ST_ATTEST_NV"
    ATTEST_COMMAND_AUDIT = "TPM_ST_ATTEST_COMMAND_AUDIT"
    ATTEST_SESSION_AUDIT = "TPM_ST_ATTEST_SESSION_AUDIT"
    ATTEST_CERTIFY = "TPM_ST_ATTEST_CERTIFY"
    ATTEST_QUOTE = "TPM_ST_ATTEST_QUOTE"
    ATTEST_TIME = "TPM_ST_ATTEST_TIME"
    ATTEST_CREATION = "TPM_ST_ATTEST_CREATION"
    CREATION = "TPM_ST_CREATION"
    VERIFIED = "TPM_ST_VERIFIED"
    AUTH_SECRET = "TPM_ST_AUTH_SECRET"
    HASHCHECK = "TPM_ST_HASHCHECK"
    AUTH_SIGNED = "TPM_ST_AUTH_SIGNED"
    FU_MANIFEST = "TPM_ST_FU_MANIFEST"


class TPM_ALG(str, Enum):
    ERROR = "TPM_ALG_ERROR"
    RSA = "TPM_ALG_RSA"
    SHA1 = "TPM_ALG_SHA1"
    HMAC = "TPM_ALG_HMAC"
    AES = "TPM_ALG_AES"
    MGF1 = "TPM_ALG_MGF1"
    KEYEDHASH = "TPM_ALG_KEYEDHASH"
    XOR = "TPM_ALG_XOR"
    SHA256 = "TPM_ALG_SHA256"
    SHA384 = "TPM_ALG_SHA384"
    SHA512 = "TPM_ALG_SHA512"
    NULL = "TPM_ALG_NULL"
    SM3_256 = "TPM_ALG_SM3_256"
    SM4 = "TPM_ALG_SM4"
    RSASSA = "TPM_ALG_RSASSA"
    RSAES = "TPM_ALG_RSAES"
    RSAPSS = "TPM_ALG_RSAPSS"
    OAEP = "TPM_ALG_OAEP"
    ECDSA = "TPM_ALG_ECDSA"
    ECDH = "TPM_ALG_ECDH"
    ECDAA = "TPM_ALG_ECDAA"
    SM2 = "TPM_ALG_SM2"
    ECSCHNORR = "TPM_ALG_ECSCHNORR"
    ECMQV = "TPM_ALG_ECMQV"
    KDF1_SP800_56A = "TPM_ALG_KDF1_SP800_56A"
    KDF2 = "TPM_ALG_KDF2"
    KDF1_SP800_108 = "TPM_ALG_KDF1_SP800_108"
    ECC = "TPM_ALG_ECC"
    SYMCIPHER = "TPM_ALG_SYMCIPHER"
    CAMELLIA = "TPM_ALG_CAMELLIA"
    CTR = "TPM_ALG_CTR"
    OFB = "TPM_ALG_OFB"
    CBC = "TPM_ALG_CBC"
    CFB = "TPM_ALG_CFB"
    ECB = "TPM_ALG_ECB"


class TPM_ECC_CURVE(str, Enum):
    NONE = "NONE"
    NIST_P192 = "NIST_P192"
    NIST_P224 = "NIST_P224"
    NIST_P256 = "NIST_P256"
    NIST_P384 = "NIST_P384"
    NIST_P521 = "NIST_P521"
    BN_P256 = "BN_P256"
    BN_P638 = "BN_P638"
    SM2_P256 = "SM2_P256"


class TPMCertInfoClockInfo:
    """
    10.11.1 TPMS_CLOCK_INFO
    """

    clock: bytes
    reset_count: int
    restart_count: int
    safe: bool

    def __init__(self, clock_info: bytes):
        self.clock = clock_info[0:8]
        self.reset_count = int.from_bytes(clock_info[8:12], "big")
        self.restart_count = int.from_bytes(clock_info[12:16], "big")
        self.safe = bool(clock_info[16])


class TPMCertInfoAttested:
    """
    10.12.3 TPMS_CERTIFY_INFO
    """

    name_alg: TPM_ALG
    name_alg_bytes: bytes
    name: bytes
    qualified_name: bytes

    def __init__(self, attested_name: bytes, qualified_name: bytes):
        self.name_alg = TPM_ALG_MAP[attested_name[0:2]]
        self.name_alg_bytes = attested_name[0:2]
        self.name = attested_name
        self.qualified_name = qualified_name


@dataclass
class TPMCertInfo:
    """
    10.12.8 TPMS_ATTEST
    """

    magic: bytes
    type: TPM_ST
    qualified_signer: bytes
    extra_data: bytes
    clock_info: TPMCertInfoClockInfo
    firmware_version: bytes
    attested: TPMCertInfoAttested


class TPMPubAreaParametersRSA:
    """
    12.2.3.5 TPMS_RSA_PARMS
    """

    symmetric: TPM_ALG
    scheme: TPM_ALG
    key_bits: bytes
    exponent: bytes

    def __init__(self, params: bytes):
        self.symmetric = TPM_ALG_MAP[params[0:2]]
        self.scheme = TPM_ALG_MAP[params[2:4]]
        self.key_bits = params[4:6]
        self.exponent = params[6:10]


class TPMPubAreaParametersECC:
    """
    12.2.3.6 TPMS_ECC_PARMS
    """

    symmetric: TPM_ALG
    scheme: TPM_ALG
    curve_id: TPM_ECC_CURVE
    kdf: TPM_ALG

    def __init__(self, params: bytes):
        self.symmetric = TPM_ALG_MAP[params[0:2]]
        self.scheme = TPM_ALG_MAP[params[2:4]]
        self.curve_id = TPM_ECC_CURVE_MAP[params[4:6]]
        self.kdf = TPM_ALG_MAP[params[6:8]]


class TPMPubAreaObjectAttributes:
    """
    8.3 TPMA_OBJECT (Object Attributes)
    """

    fixed_tpm: bool
    st_clear: bool
    fixed_parent: bool
    sensitive_data_origin: bool
    user_with_auth: bool
    admin_with_policy: bool
    no_da: bool
    encrypted_duplication: bool
    restricted: bool
    decrypt: bool
    sign_or_encrypt: bool

    def __init__(self, object_attributes: bytes):
        attrs = int.from_bytes(object_attributes[0:4], "big")
        self.fixed_tpm = attrs & (1 << 1) != 0
        self.st_clear = attrs & (1 << 2) != 0
        self.fixed_parent = attrs & (1 << 4) != 0
        self.sensitive_data_origin = attrs & (1 << 5) != 0
        self.user_with_auth = attrs & (1 << 6) != 0
        self.admin_with_policy = attrs & (1 << 7) != 0
        self.no_da = attrs & (1 << 10) != 0
        self.encrypted_duplication = attrs & (1 << 11) != 0
        self.restricted = attrs & (1 << 16) != 0
        self.decrypt = attrs & (1 << 17) != 0
        self.sign_or_encrypt = attrs & (1 << 18) != 0


class TPMPubAreaUnique:
    """
    12.2.3.2 TPMU_PUBLIC_ID
    """

    value: bytes

    def __init__(self, unique: bytes, alg_type: TPM_ALG):
        if alg_type == TPM_ALG.RSA:
            """
            As per 11.2.4.5 TPM2B_PUBLIC_KEY_RSA, extract `unique` of dynamic length
            """
            unique_length = int.from_bytes(unique[0:2], "big")
            rsa_unique = unique[2 : 2 + unique_length]
            self.value = rsa_unique
        elif alg_type == TPM_ALG.ECC:
            """
            As per 12.2.3.2 TPMU_PUBLIC_ID, `unique` is `TPMS_ECC_POINT` when `type`
            indicates ECC.

            As per 11.2.5.2 TPMS_ECC_POINT, `x` and `y` within are
            `TPM2B_ECC_PARAMETER`, which is a uint16 `size`, followed by bytes of
            the indicated length.

            `unique`, then, is structured thusly:
            [
                [x_len, ...x_bytes],
                [y_len, ...y_bytes]
            ]

            Get `unique` to this structure for easier comparison with the output from
            `decode_credential_public_key`, which will return bytes for its `x` and `y`
            and to which this value will be compared:

            [
                x_bytes,
                y_bytes,
            ]

            """
            pointer = 0
            unique_x_len = int.from_bytes(unique[0:2], "big")
            pointer += 2
            unique_x = unique[pointer : pointer + unique_x_len]
            pointer += unique_x_len
            unique_y_len = int.from_bytes(unique[pointer : pointer + 2], "big")
            pointer += 2
            unique_y = unique[pointer : pointer + unique_y_len]

            self.value = b"".join([unique_x, unique_y])
        else:
            raise InvalidTPMPubAreaStructure(
                f'Pub Area alg type "{alg_type}" was not "{TPM_ALG.RSA}" or "{TPM_ALG.ECC}"'
            )


@dataclass
class TPMPubArea:
    """
    12.2.4 TPMT_PUBLIC
    """

    type: TPM_ALG
    name_alg: TPM_ALG
    object_attributes: TPMPubAreaObjectAttributes
    auth_policy: bytes
    parameters: Union[TPMPubAreaParametersRSA, TPMPubAreaParametersECC]
    unique: TPMPubAreaUnique


@dataclass
class TPMManufacturerInfo:
    name: str
    id: str


"""
6.9 TPM_ST (Structure Tags)
"""
TPM_ST_MAP: Mapping[bytes, TPM_ST] = {
    b"\x00\xc4": TPM_ST.RSP_COMMAND,
    b"\x80\x00": TPM_ST.NULL,
    b"\x80\x01": TPM_ST.NO_SESSIONS,
    b"\x80\x02": TPM_ST.SESSIONS,
    b"\x80\x14": TPM_ST.ATTEST_NV,
    b"\x80\x15": TPM_ST.ATTEST_COMMAND_AUDIT,
    b"\x80\x16": TPM_ST.ATTEST_SESSION_AUDIT,
    b"\x80\x17": TPM_ST.ATTEST_CERTIFY,
    b"\x80\x18": TPM_ST.ATTEST_QUOTE,
    b"\x80\x19": TPM_ST.ATTEST_TIME,
    b"\x80\x1a": TPM_ST.ATTEST_CREATION,
    b"\x80\x21": TPM_ST.CREATION,
    b"\x80\x22": TPM_ST.VERIFIED,
    b"\x80\x23": TPM_ST.AUTH_SECRET,
    b"\x80\x24": TPM_ST.HASHCHECK,
    b"\x80\x25": TPM_ST.AUTH_SIGNED,
    b"\x80\x29": TPM_ST.FU_MANIFEST,
}


"""
6.3 TPM_ALG_ID
"""
TPM_ALG_MAP: Mapping[bytes, TPM_ALG] = {
    b"\x00\x00": TPM_ALG.ERROR,
    b"\x00\x01": TPM_ALG.RSA,
    b"\x00\x04": TPM_ALG.SHA1,
    b"\x00\x05": TPM_ALG.HMAC,
    b"\x00\x06": TPM_ALG.AES,
    b"\x00\x07": TPM_ALG.MGF1,
    b"\x00\x08": TPM_ALG.KEYEDHASH,
    b"\x00\x0a": TPM_ALG.XOR,
    b"\x00\x0b": TPM_ALG.SHA256,
    b"\x00\x0c": TPM_ALG.SHA384,
    b"\x00\x0d": TPM_ALG.SHA512,
    b"\x00\x10": TPM_ALG.NULL,
    b"\x00\x12": TPM_ALG.SM3_256,
    b"\x00\x13": TPM_ALG.SM4,
    b"\x00\x14": TPM_ALG.RSASSA,
    b"\x00\x15": TPM_ALG.RSAES,
    b"\x00\x16": TPM_ALG.RSAPSS,
    b"\x00\x17": TPM_ALG.OAEP,
    b"\x00\x18": TPM_ALG.ECDSA,
    b"\x00\x19": TPM_ALG.ECDH,
    b"\x00\x1a": TPM_ALG.ECDAA,
    b"\x00\x1b": TPM_ALG.SM2,
    b"\x00\x1c": TPM_ALG.ECSCHNORR,
    b"\x00\x1d": TPM_ALG.ECMQV,
    b"\x00\x20": TPM_ALG.KDF1_SP800_56A,
    b"\x00\x21": TPM_ALG.KDF2,
    b"\x00\x22": TPM_ALG.KDF1_SP800_108,
    b"\x00\x23": TPM_ALG.ECC,
    b"\x00\x25": TPM_ALG.SYMCIPHER,
    b"\x00\x26": TPM_ALG.CAMELLIA,
    b"\x00\x40": TPM_ALG.CTR,
    b"\x00\x41": TPM_ALG.OFB,
    b"\x00\x42": TPM_ALG.CBC,
    b"\x00\x43": TPM_ALG.CFB,
    b"\x00\x44": TPM_ALG.ECB,
}


"""
6.4 TPM_ECC_CURVE
"""
TPM_ECC_CURVE_MAP: Mapping[bytes, TPM_ECC_CURVE] = {
    b"\x00\x00": TPM_ECC_CURVE.NONE,
    b"\x00\x01": TPM_ECC_CURVE.NIST_P192,
    b"\x00\x02": TPM_ECC_CURVE.NIST_P224,
    b"\x00\x03": TPM_ECC_CURVE.NIST_P256,
    b"\x00\x04": TPM_ECC_CURVE.NIST_P384,
    b"\x00\x05": TPM_ECC_CURVE.NIST_P521,
    b"\x00\x10": TPM_ECC_CURVE.BN_P256,
    b"\x00\x11": TPM_ECC_CURVE.BN_P638,
    b"\x00\x20": TPM_ECC_CURVE.SM2_P256,
}


# Intentionally omit curves we can't map so a KeyError gets thrown
TPM_ECC_CURVE_COSE_CRV_MAP: Mapping[TPM_ECC_CURVE, COSECRV] = {
    TPM_ECC_CURVE.NIST_P256: COSECRV.P256,
    TPM_ECC_CURVE.NIST_P384: COSECRV.P384,
    TPM_ECC_CURVE.NIST_P521: COSECRV.P521,
    TPM_ECC_CURVE.BN_P256: COSECRV.P256,
    TPM_ECC_CURVE.SM2_P256: COSECRV.P256,
}


# Intentionally omit algs we can't map so a KeyError gets thrown
TPM_ALG_COSE_ALG_MAP: Mapping[TPM_ALG, COSEAlgorithmIdentifier] = {
    TPM_ALG.SHA256: COSEAlgorithmIdentifier.RSASSA_PSS_SHA_256,
    TPM_ALG.SHA384: COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_384,
    TPM_ALG.SHA512: COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_512,
    TPM_ALG.SHA1: COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_1,
}


# Sourced from https://trustedcomputinggroup.org/resource/vendor-id-registry/
# Latest version: https://trustedcomputinggroup.org/wp-content/uploads/TCG-TPM-Vendor-ID-Registry-Family-1.2-and-2.0-Version-1.07-Revision-0.02_pub.pdf
TPM_MANUFACTURERS: Mapping[str, TPMManufacturerInfo] = {
    "id:414D4400": TPMManufacturerInfo(name="AMD", id="AMD"),
    "id:414E5400": TPMManufacturerInfo(name="Ant Group", id="ANT"),
    "id:41544D4C": TPMManufacturerInfo(name="Atmel", id="ATML"),
    "id:4252434D": TPMManufacturerInfo(name="Broadcom", id="BRCM"),
    "id:4353434F": TPMManufacturerInfo(name="Cisco", id="CSCO"),
    "id:464C5953": TPMManufacturerInfo(name="Flyslice Technologies", id="FLYS"),
    "id:524F4343": TPMManufacturerInfo(name="Fuzhou Rockchip", id="ROCC"),
    "id:474F4F47": TPMManufacturerInfo(name="Google", id="GOOG"),
    "id:48504900": TPMManufacturerInfo(name="HPI", id="HPI"),
    "id:48504500": TPMManufacturerInfo(name="HPE", id="HPE"),
    "id:48495349": TPMManufacturerInfo(name="Huawei", id="HISI"),
    "id:49424d00": TPMManufacturerInfo(name="IBM", id="IBM"),
    "id:49465800": TPMManufacturerInfo(name="Infineon", id="IFX"),
    "id:494E5443": TPMManufacturerInfo(name="Intel", id="INTC"),
    "id:4C454E00": TPMManufacturerInfo(name="Lenovo", id="LEN"),
    "id:4D534654": TPMManufacturerInfo(name="Microsoft", id="MSFT"),
    "id:4E534D20": TPMManufacturerInfo(name="National Semiconductor", id="NSM"),
    "id:4E545A00": TPMManufacturerInfo(name="Nationz", id="NTZ"),
    "id:4E534700": TPMManufacturerInfo(name="NSING", id="NSG"),
    "id:4E544300": TPMManufacturerInfo(name="Nuvoton Technology", id="NTC"),
    "id:51434F4D": TPMManufacturerInfo(name="Qualcomm", id="QCOM"),
    "id:534D534E": TPMManufacturerInfo(name="Samsung", id="SMSN"),
    "id:53454345": TPMManufacturerInfo(name="SecEdge", id="SECE"),
    "id:534E5300": TPMManufacturerInfo(name="Sinosun", id="SNS"),
    "id:534D5343": TPMManufacturerInfo(name="SMSC", id="SMSC"),
    "id:53544D20": TPMManufacturerInfo(name="STMicroelectronics", id="STM"),
    "id:54584E00": TPMManufacturerInfo(name="Texas Instruments", id="TXN"),
    "id:57454300": TPMManufacturerInfo(name="Winbond", id="WEC"),
    "id:5345414C": TPMManufacturerInfo(name="Wisekey", id="SEAL"),
}



from unittest import TestCase

from webauthn.helpers import generate_user_handle


class TestWebAuthnGenerateUserHandle(TestCase):
    def test_generates_byte_sequence(self) -> None:
        output = generate_user_handle()

        assert type(output) == bytes
        assert len(output) == 64

    def test_generates_unique_value_each_time(self) -> None:
        output1 = generate_user_handle()
        output2 = generate_user_handle()

        assert output1 != output2



from email.mime import base
from unittest import TestCase

from webauthn.helpers import base64url_to_bytes, options_to_json
from webauthn.helpers.exceptions import InvalidJSONStructure
from webauthn.helpers.structs import (
    AuthenticatorTransport,
    PublicKeyCredentialDescriptor,
    UserVerificationRequirement,
)
from webauthn.helpers.parse_authentication_options_json import parse_authentication_options_json
from webauthn.authentication.generate_authentication_options import generate_authentication_options


class TestParseAuthenticationOptionsJSON(TestCase):
    maxDiff = None

    def test_returns_parsed_options_simple(self) -> None:
        opts = parse_authentication_options_json(
            {
                "challenge": "skxyhJljbw-ZQn-g1i87FBWeJ8_8B78oihdtSmVYaI2mArvHxI7WyTEW3gIeIRamDPlh8PJOK-ThcQc3xPNYTQ",
                "timeout": 60000,
                "rpId": "example.com",
                "allowCredentials": [],
                "userVerification": "preferred",
            }
        )

        self.assertEqual(
            opts.challenge,
            base64url_to_bytes(
                "skxyhJljbw-ZQn-g1i87FBWeJ8_8B78oihdtSmVYaI2mArvHxI7WyTEW3gIeIRamDPlh8PJOK-ThcQc3xPNYTQ"
            ),
        )
        self.assertEqual(opts.timeout, 60000)
        self.assertEqual(opts.rp_id, "example.com")
        self.assertEqual(opts.allow_credentials, [])
        self.assertEqual(opts.user_verification, UserVerificationRequirement.PREFERRED)

    def test_returns_parsed_options_full(self) -> None:
        opts = parse_authentication_options_json(
            {
                "challenge": "MTIzNDU2Nzg5MA",
                "timeout": 12000,
                "rpId": "example.com",
                "allowCredentials": [
                    {
                        "id": "MTIzNDU2Nzg5MA",
                        "type": "public-key",
                        "transports": ["internal", "hybrid"],
                    }
                ],
                "userVerification": "required",
            }
        )

        self.assertEqual(opts.challenge, base64url_to_bytes("MTIzNDU2Nzg5MA"))
        self.assertEqual(opts.timeout, 12000)
        self.assertEqual(opts.rp_id, "example.com")
        self.assertEqual(
            opts.allow_credentials,
            [
                PublicKeyCredentialDescriptor(
                    id=base64url_to_bytes("MTIzNDU2Nzg5MA"),
                    transports=[AuthenticatorTransport.INTERNAL, AuthenticatorTransport.HYBRID],
                )
            ],
        )
        self.assertEqual(opts.user_verification, UserVerificationRequirement.REQUIRED)

    def test_supports_json_string(self) -> None:
        opts = parse_authentication_options_json(
            '{"challenge": "skxyhJljbw-ZQn-g1i87FBWeJ8_8B78oihdtSmVYaI2mArvHxI7WyTEW3gIeIRamDPlh8PJOK-ThcQc3xPNYTQ", "timeout": 60000, "rpId": "example.com", "allowCredentials": [], "userVerification": "preferred"}'
        )

        self.assertEqual(
            opts.challenge,
            base64url_to_bytes(
                "skxyhJljbw-ZQn-g1i87FBWeJ8_8B78oihdtSmVYaI2mArvHxI7WyTEW3gIeIRamDPlh8PJOK-ThcQc3xPNYTQ"
            ),
        )
        self.assertEqual(opts.timeout, 60000)
        self.assertEqual(opts.rp_id, "example.com")
        self.assertEqual(opts.allow_credentials, [])
        self.assertEqual(opts.user_verification, UserVerificationRequirement.PREFERRED)

    def test_supports_options_to_json_output(self) -> None:
        """
        Test that output from `generate_authentication_options()` that's fed directly into
        `options_to_json()` gets parsed back into the original options without any changes along
        the way.
        """
        opts = generate_authentication_options(
            rp_id="example.com",
            challenge=b"1234567890",
            timeout=12000,
            allow_credentials=[
                PublicKeyCredentialDescriptor(
                    id=b"1234567890",
                    transports=[AuthenticatorTransport.INTERNAL, AuthenticatorTransport.HYBRID],
                )
            ],
            user_verification=UserVerificationRequirement.REQUIRED,
        )

        opts_json = options_to_json(opts)

        parsed_opts_json = parse_authentication_options_json(opts_json)

        self.assertEqual(parsed_opts_json.rp_id, opts.rp_id)
        self.assertEqual(parsed_opts_json.challenge, opts.challenge)
        self.assertEqual(parsed_opts_json.allow_credentials, opts.allow_credentials)
        self.assertEqual(parsed_opts_json.timeout, opts.timeout)
        self.assertEqual(parsed_opts_json.user_verification, opts.user_verification)

    def test_raises_on_non_dict_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "not a JSON object"):
            parse_authentication_options_json("[0]")

    def test_raises_on_missing_challenge(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required challenge"):
            parse_authentication_options_json({})

    def test_supports_optional_timeout(self) -> None:
        opts = parse_authentication_options_json(
            {
                "challenge": "aaa",
                "userVerification": "required",
            }
        )

        self.assertIsNone(opts.timeout)

    def test_supports_optional_rp_id(self) -> None:
        opts = parse_authentication_options_json(
            {
                "challenge": "aaa",
                "userVerification": "required",
            }
        )

        self.assertIsNone(opts.rp_id)

    def test_raises_on_missing_user_verification(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required userVerification"):
            parse_authentication_options_json(
                {
                    "challenge": "aaaa",
                }
            )

    def test_raises_on_invalid_user_verification(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "userVerification was invalid"):
            parse_authentication_options_json(
                {
                    "challenge": "aaaa",
                    "userVerification": "when_inconvenient",
                }
            )

    def test_supports_optional_allow_credentials(self) -> None:
        opts = parse_authentication_options_json(
            {
                "challenge": "aaa",
                "userVerification": "required",
            }
        )

        self.assertIsNone(opts.allow_credentials)

    def test_raises_on_allow_credentials_entry_missing_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required id"):
            parse_authentication_options_json(
                {
                    "challenge": "aaa",
                    "userVerification": "required",
                    "allowCredentials": [{}],
                }
            )

    def test_raises_on_allow_credentials_entry_invalid_transports(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "transports was not list"):
            parse_authentication_options_json(
                {
                    "challenge": "aaa",
                    "userVerification": "required",
                    "allowCredentials": [{"id": "aaaa", "transports": ""}],
                }
            )

    def test_raises_on_allow_credentials_entry_invalid_transports_entry(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "entry transports had invalid value"):
            parse_authentication_options_json(
                {
                    "challenge": "aaa",
                    "userVerification": "required",
                    "allowCredentials": [{"id": "aaaa", "transports": ["pcie"]}],
                }
            )



from unittest import TestCase

from webauthn.helpers.tpm.parse_pub_area import parse_pub_area
from webauthn.helpers.tpm.structs import TPM_ALG, TPMPubAreaParametersRSA


class TestWebAuthnTPMParsePubArea(TestCase):
    def test_properly_parses_pub_area_bytes(self) -> None:
        pub_area = b'\x00\x01\x00\x0b\x00\x06\x04r\x00 \x9d\xff\xcb\xf3l8:\xe6\x99\xfb\x98h\xdcm\xcb\x89\xd7\x158\x84\xbe(\x03\x92,\x12AX\xbf\xad"\xae\x00\x10\x00\x10\x08\x00\x00\x00\x00\x00\x01\x00\xb4\xd2\x0c\x0c\xf9-\x02\xad\x8b\xd937k\x1acE\x90\x853\x008\x08S<\xa58\x15`-{\xc0b\xf5X\xdcO\x94\x86\xf6\x18\x14\xff\xde\x01n\x15\xcf\x88\xec\xe8~\x81\xc3\xd3\xbfg\x82\x89\xe6\xbb{\xffW\xff#I!\x03\xb1\x8a\xf6\x96\x9a\x80\xa5Xj\xf8+\x9a\x01r&\xbc\\\xae\x1dU,\x86\xc3q\'Uks\xe5K/\x14Sw\x01\xfaE\x9d\x81\xfcJWb\x80c\x1b\xc4\xf3\xc3\x1fd\x9f"\xfd\xce[8z\xe0\xea\xeb\x18\x81\xd3\x8fc\xab\xfa\x84\x8d\x840\\\x8fS\x1b>#R\xb1\x91\xd0\xa6\xf6&R\xee\xc56\x18T%"\xa5\xa23i\x96\xb0\xa6{\x82\xc8@\x14\xb0T<N\xb2\xd2\xceZ2^g\x06\xd27X\\\xd8@Y\x01\x1a\xc9\x01\x8f\xafd\x98\x96\x06ntE\xc7\x89\x88\xa5\xa6\xebVU\xdf\xa9\xe3\xe0\x9e`,\xa6\xc2\x03\xc5\xc5w\x08vZ|\x08K\xd9D\xa2\xb7@\xb2^F\xb2\xa2(\x0e\xae\xe4\xa6\xa9\xd3\xe0I1R;A\xd7'

        output = parse_pub_area(pub_area)

        assert output.type == TPM_ALG.RSA
        assert output.name_alg == TPM_ALG.SHA256
        assert (
            output.auth_policy
            == b'\x9d\xff\xcb\xf3l8:\xe6\x99\xfb\x98h\xdcm\xcb\x89\xd7\x158\x84\xbe(\x03\x92,\x12AX\xbf\xad"\xae'
        )
        assert (
            output.unique.value
            == b'\xb4\xd2\x0c\x0c\xf9-\x02\xad\x8b\xd937k\x1acE\x90\x853\x008\x08S<\xa58\x15`-{\xc0b\xf5X\xdcO\x94\x86\xf6\x18\x14\xff\xde\x01n\x15\xcf\x88\xec\xe8~\x81\xc3\xd3\xbfg\x82\x89\xe6\xbb{\xffW\xff#I!\x03\xb1\x8a\xf6\x96\x9a\x80\xa5Xj\xf8+\x9a\x01r&\xbc\\\xae\x1dU,\x86\xc3q\'Uks\xe5K/\x14Sw\x01\xfaE\x9d\x81\xfcJWb\x80c\x1b\xc4\xf3\xc3\x1fd\x9f"\xfd\xce[8z\xe0\xea\xeb\x18\x81\xd3\x8fc\xab\xfa\x84\x8d\x840\\\x8fS\x1b>#R\xb1\x91\xd0\xa6\xf6&R\xee\xc56\x18T%"\xa5\xa23i\x96\xb0\xa6{\x82\xc8@\x14\xb0T<N\xb2\xd2\xceZ2^g\x06\xd27X\\\xd8@Y\x01\x1a\xc9\x01\x8f\xafd\x98\x96\x06ntE\xc7\x89\x88\xa5\xa6\xebVU\xdf\xa9\xe3\xe0\x9e`,\xa6\xc2\x03\xc5\xc5w\x08vZ|\x08K\xd9D\xa2\xb7@\xb2^F\xb2\xa2(\x0e\xae\xe4\xa6\xa9\xd3\xe0I1R;A\xd7'
        )
        # Object Attributes
        assert output.object_attributes.fixed_tpm is True
        assert output.object_attributes.st_clear is False
        assert output.object_attributes.fixed_parent is True
        assert output.object_attributes.sensitive_data_origin is True
        assert output.object_attributes.user_with_auth is True
        assert output.object_attributes.admin_with_policy is False
        assert output.object_attributes.no_da is True
        assert output.object_attributes.encrypted_duplication is False
        assert output.object_attributes.restricted is False
        assert output.object_attributes.decrypt is True
        assert output.object_attributes.sign_or_encrypt is True
        # Parameters
        assert output.parameters.symmetric == TPM_ALG.NULL
        assert output.parameters.scheme == TPM_ALG.NULL
        assert isinstance(output.parameters, TPMPubAreaParametersRSA)
        assert output.parameters.key_bits == b"\x08\x00"
        assert output.parameters.exponent == b"\x00\x00\x00\x00"



from .structs import TPM_MANUFACTURERS, TPMManufacturerInfo


def map_tpm_manufacturer_id(id: str) -> TPMManufacturerInfo:
    """
    Map a TPM manufacturer's hex ID to a manufacturer's assigned name and ASCII identifier

    Args:
        - `id`: A TPM manufacturer ID string like `"id:FFFFFFFF"`
          (a.k.a. oid "2.23.133.2.1" in SubjectAlternativeName extension)

    Returns:
        An instance of `TPMManufacturerInfo`

    Raises:
        `KeyError` on unrecognized TPM manufacturer ID
    """
    return TPM_MANUFACTURERS[id]



from unittest import TestCase

from webauthn.helpers.bytes_to_base64url import bytes_to_base64url


class TestWebAuthnBytesToBase64URL(TestCase):
    def test_converts_buffer_to_base64url_string(self) -> None:
        output = bytes_to_base64url(bytes([1, 2, 3, 4, 5]))

        assert output == "AQIDBAU"
